/*
 * Polygon_RayTrace
 * Copylight (C) 2013 mocchi
 * mocchi_2003@yahoo.co.jp
 * License: Boost ver.1
 */

#define USE_EMBREE

#include <cstdio>
#include <cmath>
#include <algorithm>
#include <chrono>

#include "opennurbs.h"
#include "thpool.h"
#include "mist/thread.h"
#include "yyjson_pp.h"
#include "rply.h"
#include "gd.h"

#define TINYEXR_IMPLEMENTATION
#define TINYEXR_USE_MINIZ 0
#ifdef min
#undef min
#undef max
#endif
#include "tinyexr.h"

#include "PhisicalProperties.h"
#include "randomizer.h"

#include <windows.h>

#define RAY_IOTA_PROGRESS 0.005
#define TRACE_TERMINAL_LENGTH 10.0
#define MAX_INTERSECTION_COUNT 5000

#ifdef USE_EMBREE
#include "embree3/rtcore.h"
#endif

float get_ieee754(ON__UINT8 p[4]){
	return *reinterpret_cast<float *>(p);
}

// Ql: http://ja.wikipedia.org/wiki/Standard_Triangulated_Language
bool ConvertFromBinarySTL(ON_SimpleArray<ON__UINT8> &stl, double scale, const ON_3dPoint &position, ON_Mesh &mesh){
	mesh.Destroy();
	if (stl.Count() < 84) return false;
	ON__UINT32 num_facets =
		static_cast<ON__UINT32>(stl[80]) + static_cast<ON__UINT32>(stl[81]) * 256 +
		static_cast<ON__UINT32>(stl[82]) * 65536 + static_cast<ON__UINT32>(stl[83]) * 16777216;

	if (stl.Count() < 84 + static_cast<int>(num_facets) * 50 - 2) return false;

	mesh.DestroyDoublePrecisionVertices();
	mesh.SetSinglePrecisionVerticesAsValid();
	ON__UINT8 *p = stl.First() + 84;
	for (ON__UINT32 i = 0; i < num_facets; ++i, p += 50){
		int vcnt = mesh.m_V.Count();
		mesh.m_FN.AppendNew().Set(get_ieee754(p   ), get_ieee754(p+ 4), get_ieee754(p+ 8));
		mesh.m_FN.Last()->Unitize();
		mesh.m_V.AppendNew().Set(get_ieee754(p+12), get_ieee754(p+16), get_ieee754(p+20));
		mesh.m_V.AppendNew().Set(get_ieee754(p+24), get_ieee754(p+28), get_ieee754(p+32));
		mesh.m_V.AppendNew().Set(get_ieee754(p+36), get_ieee754(p+40), get_ieee754(p+44));
		mesh.SetTriangle(mesh.m_F.Count(), vcnt, vcnt+1, vcnt+2);
	}
	mesh.CombineIdenticalVertices();
	for (int i = 0; i < mesh.m_V.Count(); ++i) {
		mesh.m_V[i] *= scale;
		mesh.m_V[i] += ON_3fPoint(position);
	}
	mesh.Compact();
	return true;
}

bool ConvertFromPLY(const char *filename, double scale, const ON_3dPoint &position, ON_Mesh &mesh) {
	bool rc = false;
	auto ply = ply_open(filename, [](p_ply ply, const char *msg) { }, 0, nullptr);
	if (!ply) return false;
	if (!ply_read_header(ply)) goto END;
	mesh.Destroy();
	mesh.DestroyDoublePrecisionVertices();
	mesh.SetSinglePrecisionVerticesAsValid();

	{
		auto add_vertex = [](p_ply_argument arg) {
			void *pdata;
			long xyz;
			ply_get_argument_user_data(arg, &pdata, &xyz);
			ON_Mesh *mesh = static_cast<ON_Mesh *>(pdata);
			long index;
			p_ply_element elm;
			ply_get_argument_element(arg, &elm, &index);
			mesh->m_V[index][xyz] = ply_get_argument_value(arg);
			return 1;
		};
		long nvertices = ply_set_read_cb(ply, "vertex", "x", add_vertex, &mesh, 0);
		ply_set_read_cb(ply, "vertex", "y", add_vertex, &mesh, 1);
		ply_set_read_cb(ply, "vertex", "z", add_vertex, &mesh, 2);
		mesh.m_V.SetCapacity(nvertices);
		mesh.m_V.SetCount(nvertices);

		auto add_face = [](p_ply_argument arg) {
			void *pdata;
			ply_get_argument_user_data(arg, &pdata, nullptr);
			ON_Mesh *mesh = static_cast<ON_Mesh *>(pdata);
			long index;
			p_ply_element elm;
			ply_get_argument_element(arg, &elm, &index);
			long length, vi;
			ply_get_argument_property(arg, nullptr, &length, &vi);
			if (vi < 0) return 1;
			mesh->m_F[index].vi[vi] = static_cast<int>(ply_get_argument_value(arg));
			if (vi == 2) {
				mesh->m_F[index].vi[3] = mesh->m_F[index].vi[2];
			}
			return 1;
		};
		long ntriangles = ply_set_read_cb(ply, "face", "vertex_indices", add_face, &mesh, 0);
		mesh.m_F.SetCapacity(ntriangles);
		mesh.m_F.SetCount(ntriangles);
		if (!ply_read(ply)) goto END;
	}
	mesh.CombineIdenticalVertices();
	mesh.ComputeVertexNormals();
	for (int i = 0; i < mesh.m_V.Count(); ++i) {
		mesh.m_V[i] *= scale;
		mesh.m_V[i] += ON_3fPoint(position);
	}
	mesh.Compact();

	rc = true;
END:
	ply_close(ply);
	return true;
}

bool ReadFile(const char *filename, ON_SimpleArray<ON__UINT8> &data){
	FILE *fp = std::fopen(filename, "rb");
	if (!fp) return false;
	std::fseek(fp, 0, SEEK_END);
	data.SetCapacity(static_cast<int>(std::ftell(fp)));
	data.SetCount(data.Capacity());
	std::rewind(fp);
	std::fread(data, 1, data.Count(), fp);
	std::fclose(fp);
	return true;
}

struct MeshRayIntersection{
	std::unique_ptr<RTCIntersectContext> context;
	RTCScene *scene;
	ON_Mesh *mesh;

	void Initialize(ON_Mesh *mesh_, RTCScene *scene_){
		scene = scene_;
		mesh = mesh_;
		context.reset(new RTCIntersectContext());
		rtcInitIntersectContext(context.get());
	}

	// vZ
	ON_RTreeCapsule cc;
	int mesh_idx;
	int face_idx;
	double u, v;
	bool RayIntersection(const ON_3dRay &ray){

		RTCRayHit rayhit;
		rayhit.ray.org_x = ray.m_P.x;
		rayhit.ray.org_y = ray.m_P.y;
		rayhit.ray.org_z = ray.m_P.z;
		rayhit.ray.dir_x = ray.m_V.x;
		rayhit.ray.dir_y = ray.m_V.y;
		rayhit.ray.dir_z = ray.m_V.z;
		rayhit.ray.tnear = 0;
		rayhit.ray.tfar = std::numeric_limits<float>::infinity();
		rayhit.ray.mask = -1;
		rayhit.ray.flags = 0;
		rayhit.hit.geomID = RTC_INVALID_GEOMETRY_ID;
		rayhit.hit.instID[0] = RTC_INVALID_GEOMETRY_ID;

		rtcIntersect1(*scene, context.get(), &rayhit);

		if (rayhit.hit.geomID != RTC_INVALID_GEOMETRY_ID){
			mesh_idx = rayhit.hit.geomID;
			face_idx = rayhit.hit.primID;
			ON_3dPoint newpt = ray.m_P + ray.m_V * rayhit.ray.tfar;
			u = rayhit.hit.u;
			v = rayhit.hit.v;
			for (int h = 0; h < 3; ++h) cc.m_point[1][h] = newpt[h];
			return true;
		}else return false;
	}
	
};

struct Environment {
	struct fRGB {
		float r, g, b;
	};
	ON_ClassArray<fRGB> image;
	int width, height;
	ON_3dVector zenith, center, equator;
	Environment(yyjson_val *enp) {
		width = height = 0;
		if (!enp) return;
		auto &env = *enp;

		auto zenith_p = env["zenith_dir"];
		if (zenith_p && zenith_p->arr_size() == 3) {
			zenith.Set(yyjson_get_real_cast((*zenith_p)[0ULL]), yyjson_get_real_cast((*zenith_p)[1]), yyjson_get_real_cast((*zenith_p)[2]));
		} else {
			zenith.Set(0, 0, 0);
		}

		auto center_p = env["center_dir"];
		if (center_p && center_p->arr_size() == 3) {
			center.Set(yyjson_get_real_cast((*center_p)[0ULL]), yyjson_get_real_cast((*center_p)[1]), yyjson_get_real_cast((*center_p)[2]));
		} else {
			center.Set(0, 0, 0);
		}

		equator = ON_CrossProduct(center, zenith);
		zenith.Unitize();
		center.Unitize();
		equator.Unitize();
		double multiplier = yyjson_get_real_cast(env["multiplier"]);
		LoadEXR(yyjson_get_str(env["path"]), multiplier);
	}

	const fRGB &operator ()(ON_3dVector &ray_dir) {
		//                      Z:zenith
		//         dr(dx,dy,dz)  |    X:center
		//        (0,ty,dz) *-_  |   / 
		//             _  +-:--~-+dz/
		//              ~-: :    | /
		//             u( :~*-_  |/
		// Y:equator    --+----~-+--------
		// YZ v[deg]񂵂ɁA X u[deg] 񂷂Ƃ dr ɂȂB
		// (1) 2]̃xNg̒^2: dx^2 + dy^2 + dz^2
		// (2) 1]̃xNg̒^2: ty^2 + dz^2
		// (1) = (2) Aty^2 = dx^2 + dy^2 => ty = sqrt(dx^2 + dy^2)
		// tan(v) = dz / ty
		// tan(u) = dy / dx

		double dz = ON_DotProduct(zenith, ray_dir);
		double dy = ON_DotProduct(equator, ray_dir);
		double dx = ON_DotProduct(center, ray_dir);
		double ty = std::sqrt(dx * dx + dy * dy);

		double v_rad = (ty < ON_ZERO_TOLERANCE) ? 0 : std::atan2(dz, ty);
		double u_rad = (std::abs(dx) < ON_ZERO_TOLERANCE) ? 0 : std::atan2(dy, dx);

		// u_rad : [-PI to PI], v_rad : [-PI/2 to PI/2]
		// (0,0) ̂Ƃ㉺E, (X,PI/2) ̂Ƃ㒆
		int px = (u_rad / (2.0 * ON_PI) + 0.5) * static_cast<double>(width);
		if (px < 0) px = 0;
		else if (px >= width) px = width;
		int py = (-v_rad / ON_PI + 0.5) * static_cast<double>(height);
		if (py < 0) py = 0;
		else if (py >= height) py = height;
		return (*this)(px, py);
	}

private:
	fRGB &operator ()(int px, int py) {
		return image[py*width + px];
	}
	fRGB &operator[] (size_t idx) {
		return image[idx];
	}

	void resize(int width_, int height_) {
		width = width_;
		height = height_;
		image.SetCapacity(width * height);
		image.SetCount(image.Capacity());
	}


	void LoadEXR(const char *path_exr, double multiplier) {
		if (!path_exr) return;
		std::fprintf(stderr, "  %s\n", path_exr);
		image.Empty();
		ON_SimpleArray<uint8_t> exrdata;
		{
			std::unique_ptr<FILE, decltype(&std::fclose)> fp(std::fopen(path_exr, "rb"), std::fclose);
			if (!fp.get()) return;
			std::fseek(fp.get(), 0, SEEK_END);
			exrdata.SetCapacity(std::ftell(fp.get()));
			exrdata.SetCount(exrdata.Capacity());
			std::fseek(fp.get(), 0, SEEK_SET);
			std::fread(exrdata.First(), exrdata.Count(), 1, fp.get());
		}

		EXRVersion exr_version;
		EXRImage exr_image;
		EXRHeader exr_header;
		const char *err = 0;
		InitEXRHeader(&exr_header);
		std::unique_ptr<EXRHeader, decltype(&::FreeEXRHeader)> deleter_exr_header(&exr_header, ::FreeEXRHeader);
		InitEXRImage(&exr_image);
		std::unique_ptr<EXRImage, decltype(&::FreeEXRImage)> deleter_exr_image(&exr_image, ::FreeEXRImage);

		{
			int ret = ParseEXRVersionFromMemory(&exr_version, exrdata, exrdata.Count());
			if (ret != TINYEXR_SUCCESS) return;
			// "Failed to open EXR file or read version info from EXR file."

			if (exr_version.multipart || exr_version.non_image) return;
			// "Loading multipart or DeepImage is not supported  in LoadEXR() API"
		}

		{
			int ret = ParseEXRHeaderFromMemory(&exr_header, &exr_version, exrdata, exrdata.Count(), &err);
			if (ret != TINYEXR_SUCCESS) return;
		}

		// Read HALF channel as FLOAT.
		for (int i = 0; i < exr_header.num_channels; i++) {
			if (exr_header.pixel_types[i] == TINYEXR_PIXELTYPE_HALF) {
				exr_header.requested_pixel_types[i] = TINYEXR_PIXELTYPE_FLOAT;
			}
		}

		double xr, yr, xg, yg, xb, yb, xw, yw;
		bool chromacity_predifined;
		OpenEXR_ReadChromacity(exr_header, xr, yr, xg, yg, xb, yb, xw, yw, &chromacity_predifined);

		{
			int ret = LoadEXRImageFromMemory(&exr_image, &exr_header, exrdata, exrdata.Count(), &err);
			if (ret != TINYEXR_SUCCESS) return;
		}
		this->resize(exr_image.width, exr_image.height);

		// Get RGBA channel Index
		int idxR = -1, idxG = -1, idxB = -1, idxA = -1;
		for (int c = 0; c < exr_header.num_channels; c++) {
			if (std::strcmp(exr_header.channels[c].name, "R") == 0) {
				idxR = c;
			}
			else if (std::strcmp(exr_header.channels[c].name, "G") == 0) {
				idxG = c;
			}
			else if (std::strcmp(exr_header.channels[c].name, "B") == 0) {
				idxB = c;
			}
			else if (std::strcmp(exr_header.channels[c].name, "A") == 0) {
				idxA = c;
			}
		}

		if (exr_header.num_channels == 1) {
			// Grayscale channel only.

			if (exr_header.tiled) {
				for (int it = 0; it < exr_image.num_tiles; it++) {
					for (int j = 0; j < exr_header.tile_size_y; j++) {
						for (int i = 0; i < exr_header.tile_size_x; i++) {
							const int ii = exr_image.tiles[it].offset_x * exr_header.tile_size_x + i;
							const int jj = exr_image.tiles[it].offset_y * exr_header.tile_size_y + j;
							const int idx = ii + jj * exr_image.width;

							// out of region check.
							if (ii >= exr_image.width) continue;
							if (jj >= exr_image.height) continue;

							const int srcIdx = i + j * exr_header.tile_size_x;
							unsigned char **src = exr_image.tiles[it].images;
							float gray = reinterpret_cast<float **>(src)[0][srcIdx];
							auto &rgb = (*this)(ii, jj);
							rgb.r = rgb.g = rgb.b = gray;
						}
					}
				}
			}
			else {
				for (int i = 0; i < exr_image.width * exr_image.height; i++) {
					auto &rgb = (*this)[i];
					rgb.r = rgb.g = rgb.b = reinterpret_cast<float **>(exr_image.images)[0][i];
				}
			}
		}
		else {
			if (idxR == -1 || idxG == -1 || idxB == -1) return;

			if (exr_header.tiled) {
				for (int it = 0; it < exr_image.num_tiles; it++) {
					for (int j = 0; j < exr_header.tile_size_y; j++) {
						for (int i = 0; i < exr_header.tile_size_x; i++) {
							const int ii = exr_image.tiles[it].offset_x * exr_header.tile_size_x + i;
							const int jj = exr_image.tiles[it].offset_y * exr_header.tile_size_y + j;
							const int idx = ii + jj * exr_image.width;

							// out of region check.
							if (ii >= exr_image.width) continue;
							if (jj >= exr_image.height) continue;

							const int srcIdx = i + j * exr_header.tile_size_x;
							unsigned char **src = exr_image.tiles[it].images;
							auto &rgb = (*this)(ii, jj);
							rgb.r = reinterpret_cast<float **>(src)[idxR][srcIdx];
							rgb.g = reinterpret_cast<float **>(src)[idxG][srcIdx];
							rgb.b = reinterpret_cast<float **>(src)[idxB][srcIdx];
						}
					}
				}
			}
			else {
				for (int i = 0; i < exr_image.width * exr_image.height; i++) {
					auto &rgb = (*this)[i];
					rgb.r = reinterpret_cast<float **>(exr_image.images)[idxR][i];
					rgb.g = reinterpret_cast<float **>(exr_image.images)[idxG][i];
					rgb.b = reinterpret_cast<float **>(exr_image.images)[idxB][i];
				}
			}
		}

		fRGB rgb_max = { 0,0,0 };
		for (int i = 0; i < exr_image.width * exr_image.height; i++) {
			auto &rgb = (*this)[i];
			rgb.r *= multiplier;
			rgb.g *= multiplier;
			rgb.b *= multiplier;
			if (rgb_max.r < rgb.r) rgb_max.r = rgb.r;
			if (rgb_max.g < rgb.g) rgb_max.g = rgb.g;
			if (rgb_max.b < rgb.b) rgb_max.b = rgb.b;
		}
		std::fprintf(stderr, "  rgb_max:(%f, %f, %f)\n", rgb_max.r, rgb_max.g, rgb_max.b);

	}
	bool OpenEXR_ReadChromacity(const EXRHeader &exr_header, double &xr, double &yr, double &xg, double &yg, double &xb, double &yb, double &xw, double &yw, bool *pchromacities_defined) {
		bool rc = false;
		const char *err = 0;

		auto get_float = [](unsigned char *p) {
			float v = *reinterpret_cast<float *>(p);
			tinyexr::swap4(&v);
			return v;
		};

		bool chromacities_defined = false;
		for (int i = 0; i < exr_header.num_custom_attributes; ++i) {
			EXRAttribute &attr = exr_header.custom_attributes[i];
			if (std::strcmp(attr.name, "chromaticities") != 0) continue;
			chromacities_defined = true;
			// `gGfBAƉ
			xr = get_float(attr.value);
			yr = get_float(attr.value + 4);
			xg = get_float(attr.value + 8);
			yg = get_float(attr.value + 12);
			xb = get_float(attr.value + 16);
			yb = get_float(attr.value + 20);
			xw = get_float(attr.value + 24);
			yw = get_float(attr.value + 28);
			break;
		}
		if (!chromacities_defined) {
			// https://openexr.readthedocs.io/en/latest/TechnicalIntroduction.html
			// Recommendations - RGB Color
			xr = 0.6400, yr = 0.3300;
			xg = 0.3000, yg = 0.6000;
			xb = 0.1500, yb = 0.0600;
			xw = 0.3127, yw = 0.3290;
		}
		if (pchromacities_defined) *pchromacities_defined = chromacities_defined;

		rc = true;
		return rc;
	}
};

void read_3real(yyjson_val *j_cmr, const char *name, double *dest) {
	auto *j_arr = (*j_cmr)[name];
	if (j_arr && j_arr->arr_size() == 3) {
		dest[0] = yyjson_get_real_cast((*j_arr)[0ULL]);
		dest[1] = yyjson_get_real_cast((*j_arr)[1]);
		dest[2] = yyjson_get_real_cast((*j_arr)[2]);
	} else {
		dest[0] = dest[1] = dest[2] = 0;
	}
};

struct Cameras {
	struct Camera {
		enum ProjectionMode {
			Parallel, Perspective
		}proj_mode;
		ON_3dPoint origin;
		ON_3dVector eye, vert, horz;
		ON_Plane pln;
		int pixel_width, pixel_height;
		double horz_range, vert_range;
		int pass;
		ON_String output_filename;
	};
	ON_SimpleArray<Camera> cameras;
	Cameras(yyjson_val *j_cmrs) {
		if (!j_cmrs) return;

		for (size_t i = 0; i < j_cmrs->arr_size(); ++i) {
			Camera &cmr = cameras.AppendNew();
			auto *j_cmr = (*j_cmrs)[i];
#if 0
			auto read_3real = [j_cmr](const char *name, double *dest) {
				auto *j_arr = (*j_cmr)[name];
				if (j_arr && j_arr->arr_size() == 3) {
					dest[0] = yyjson_get_real_cast((*j_arr)[0ULL]);
					dest[1] = yyjson_get_real_cast((*j_arr)[1]);
					dest[2] = yyjson_get_real_cast((*j_arr)[2]);
				} else {
					dest[0] = dest[1] = dest[2] = 0;
				}
			};
#endif
			read_3real(j_cmr, "origin", static_cast<double *>(cmr.origin));
			read_3real(j_cmr, "eye_dir", static_cast<double *>(cmr.eye));
			read_3real(j_cmr, "horz_dir", static_cast<double *>(cmr.horz));
			read_3real(j_cmr, "vert_dir", static_cast<double *>(cmr.vert));
			cmr.pixel_width = yyjson_get_int((*j_cmr)["pixel_width"]);
			cmr.pixel_height = yyjson_get_int((*j_cmr)["pixel_height"]);
			cmr.horz_range = yyjson_get_real_cast((*j_cmr)["horz_range"]);
			cmr.vert_range = yyjson_get_real_cast((*j_cmr)["vert_range"]);
			cmr.output_filename = yyjson_get_str((*j_cmr)["output_filename"]);
			double far_ = yyjson_get_real_cast((*j_cmr)["far"]);
			cmr.pass = yyjson_get_int((*j_cmr)["pass"]);

			const char *prj_mode = yyjson_get_str((*j_cmr)["projection_mode"]);
			if (prj_mode && std::strcmp(prj_mode, "parallel") == 0) {
				cmr.proj_mode = Camera::Parallel;
				cmr.pln.CreateFromFrame(cmr.origin, cmr.horz, cmr.vert);
			} else {
				cmr.proj_mode = Camera::Perspective;
				cmr.pln.CreateFromFrame(cmr.origin + cmr.eye * far_, cmr.horz, cmr.vert);
			}
		}
	}
};

struct CommonInfo{
	LightSources *light_src;
	Materials *materials;
	Environment *environment;
	int cnt_10;
	pthread_mutex_t mtx_read, mtx_write;

	CommonInfo() {
		::pthread_mutex_init(&mtx_read, 0);
		::pthread_mutex_init(&mtx_write, 0);
	}
	~CommonInfo() {
		::pthread_mutex_destroy(&mtx_read);
		::pthread_mutex_destroy(&mtx_write);
	}

	// read
	struct Scene {
		ON_Mesh *mesh;
		ON_3dPoint model_center;
		double rough_radius;

		RTCDevice device;
		RTCScene scene;

		static void errorFunction(void* userPtr, enum RTCError error, const char* str) {
			std::printf("error %d: %s\n", error, str);
		}

		void Initialize(ON_Mesh *mesh_) {
			mesh = mesh_;
			ON_BoundingBox tbb;
			mesh_->GetTightBoundingBox(tbb);
			rough_radius = tbb.Diagonal().Length() * 0.5;
			model_center = tbb.Center();

			device = rtcNewDevice(0);
			if (!device) {
				std::printf("error %d: cannot create device\n", rtcGetDeviceError(0));
				return;
			}
			rtcSetDeviceErrorFunction(device, errorFunction, 0);

			scene = rtcNewScene(device);

			RTCGeometry geom = rtcNewGeometry(device, RTC_GEOMETRY_TYPE_TRIANGLE);
			float* vertices = static_cast<float*>(
				rtcSetNewGeometryBuffer(geom, RTC_BUFFER_TYPE_VERTEX, 0, RTC_FORMAT_FLOAT3, 3 * sizeof(float), mesh_->VertexCount())
				);
			unsigned int* indices = static_cast<unsigned int*>(
				rtcSetNewGeometryBuffer(geom, RTC_BUFFER_TYPE_INDEX, 0, RTC_FORMAT_UINT3, 3 * sizeof(unsigned), mesh_->FaceCount())
				);

			for (int i = 0, i3 = 0; i < mesh_->VertexCount(); ++i, i3 += 3) {
				ON_3dPoint &pt = mesh_->Vertex(i);
				vertices[i3] = pt.x;
				vertices[i3 + 1] = pt.y;
				vertices[i3 + 2] = pt.z;
			}
			for (int i = 0, i3 = 0; i < mesh_->FaceCount(); ++i, i3 += 3) {
				ON_MeshFace &f = mesh_->m_F[i];
				indices[i3] = f.vi[0];
				indices[i3 + 1] = f.vi[1];
				indices[i3 + 2] = f.vi[2];
			}

			rtcCommitGeometry(geom);

			rtcAttachGeometry(scene, geom);
			rtcReleaseGeometry(geom);
			rtcCommitScene(scene);
		}
		~Scene() {
			if (!device) return;
			rtcReleaseScene(scene);
			rtcReleaseDevice(device);
		}
	}scene;

	int ray_cursor;
	ON_SimpleArray<unsigned int> *shapeidx2fidx;
	ON_SimpleArray<int> *shape2matidx;
	ON_SimpleArray<FaceNormalDirectionMode> *shape2fndm;

	// write
	ON_ClassArray<ON_ClassArray<ON_3dRay> > raies_last;
	ON_ClassArray<ON_SimpleArray<double> > flux_last;
	ON_ClassArray<ON_Polyline> traces;
};

int RayTrace(const ON_3dRay &ray_, double flux, MeshRayIntersection &mri, CommonInfo *ci, xorshift_rnd_32bit &rnd, ON_3dRay &ray_o, double power[3], ON_Polyline *trace, bool &error){
	error = false;
	ON_3dRay ray = ray_;
	if (trace) trace->Append(ray.m_P);
	bool is_inside = false;
	bool absorbed = false;
	ON_Mesh &cshape = *mri.mesh;
	int cnt = 0;

	while (mri.RayIntersection(ray)) {
		++cnt;
		// ʔԍ`ԍ擾
		int shape_idx = static_cast<int>(
			std::upper_bound(ci->shapeidx2fidx->First(),
							 ci->shapeidx2fidx->Last() + 1, static_cast<unsigned int>(mri.face_idx)) - ci->shapeidx2fidx->First()
			) - 1;

		// flat shading
		auto get_flat_nrm = [&]() {
			return cshape.m_FN[mri.face_idx];
		};

		int midx = (*ci->shape2matidx)[shape_idx];
		// phong shading
		auto get_phong_nrm = [&]() {
			int fi = mri.face_idx;
			return
				cshape.m_N[cshape.m_F[fi].vi[0]] * (1.0 - (mri.u + mri.v)) +
				cshape.m_N[cshape.m_F[fi].vi[1]] * mri.u +
				cshape.m_N[cshape.m_F[fi].vi[2]] * mri.v;
		};
		ON_3dVector phong_nrm = get_phong_nrm(), flat_nrm = get_flat_nrm();

		phong_nrm.Unitize();
		flat_nrm.Unitize();

		// ގ̎͋zWKp
		if (is_inside) {
			double dist = ray.m_P.DistanceTo(ON_3dPoint(mri.cc.m_point[1]));
			ci->materials->VolumeAttenuate(midx, 3, power, dist);
		}

		ON_3dVector emit_dir;
		bool is_inside_prev = is_inside;
		FaceNormalDirectionMode fndm = (*ci->shape2fndm)[shape_idx];
		bool rc = ci->materials->CalcBSDF(midx, fndm, phong_nrm, ray.m_V, is_inside, rnd, 3, power, emit_dir);
		if (!rc) {
			error = true;
			break;
		}
		{
			// ˂Ȃ̂ɓ˂ĂA܂́A߂Ȃ̂ɓ}ɂƂǂ܂P[X̃`FbNB
			// phong_nrm  flat_nrm ̂ɂAȗ̑傫Ƃ diffuse Ő󂢊pxւ̔ˁAŋN₷B

			if (ON_DotProduct(phong_nrm, flat_nrm) < 0) flat_nrm.Reverse();
			// ̎_ flat_nrm  incident ƔΌɂȂĂB

			if (!emit_dir.IsZero()) {
				bool new_dir_is_reflection = (ON_DotProduct(flat_nrm, emit_dir) > 0);
				if ((new_dir_is_reflection && (is_inside != is_inside_prev)) || (!new_dir_is_reflection && (is_inside == is_inside_prev))) {
					error = true;
					break;
				}
			}
		}
		ray.m_V = emit_dir;
		if (ray.m_V.IsZero() || (power[0] == 0 && power[1] == 0 && power[2] == 0)) {
			absorbed = true;
			break;
		}
		if (trace) trace->Append(mri.cc.m_point[1]);
		ray.m_P = ON_3dPoint(mri.cc.m_point[1]) + ray.m_V * RAY_IOTA_PROGRESS;
		if (cnt >= MAX_INTERSECTION_COUNT) {
			error = true;
			break;
		}
	}

	if (trace && !absorbed) trace->Append(ray.m_P + ray.m_V * TRACE_TERMINAL_LENGTH);
	ray_o = ray;
	return cnt;
}

int main(int argc, char *argv[]){
	if (argc == 1){
		return 0;
	}
	std::unique_ptr<yyjson_doc, decltype(&yyjson_doc_free)> args_doc(0, yyjson_doc_free);
	{
		FILE *fp = std::fopen(argv[1], "rb");
		std::fseek(fp, 0, SEEK_END);
		ON_SimpleArray<char> settingfile(std::ftell(fp));
		std::rewind(fp);
		settingfile.SetCount(settingfile.Capacity());
		std::fread(settingfile, 1, settingfile.Count(), fp);
		std::fclose(fp);
		args_doc.reset(yyjson_read(settingfile, settingfile.Count(), YYJSON_READ_NOFLAG));
	}
	yyjson_val *args_p = yyjson_doc_get_root(args_doc.get());

	ON_ClassArray<ON_Mesh> shapes;
	ON_SimpleArray<int> shape2matidx;
	ON_SimpleArray<FaceNormalDirectionMode> shape2fndm;
	// SẴbV̍
	ON_Mesh cshape;

	// ̖ʔԍ獇Ǒ`ԍ擾邽߂ɎgpB
	ON_SimpleArray<unsigned int> shapeidx2fidx;

	std::fprintf(stderr, "Reading shapes.\n");
	auto shapes_p = (*args_p)["shapes"];
	if (yyjson_is_arr(shapes_p)){
		for (size_t k = 0; k < shapes_p->arr_size(); ++k){
			ON_Mesh &shape = shapes.AppendNew();
			auto shape_p = (*shapes_p)[k];
			const char *filename = yyjson_get_str((*shape_p)["filename"]);
			if (filename[0] == '\0') continue;
			const char *face_direction = yyjson_get_str((*shape_p)["face_direction"]);

			double scale = yyjson_get_real_cast((*shape_p)["scale"]);
			ON_3dPoint position;
			read_3real(shape_p, "position", position);

			FaceNormalDirectionMode fndm = FaceNormalDirectionMode::AUTO;
			if (std::strcmp(face_direction, "outer") == 0) {
				fndm = FaceNormalDirectionMode::OUTER;
			} else if (std::strcmp(face_direction, "inner") == 0) {
				fndm = FaceNormalDirectionMode::INNER;
			}
			shape2fndm.Append(fndm);
			std::fprintf(stderr, "  %s\n", filename);

			if (std::strstr(filename, ".3dm") != 0) {
				ONX_Model model;
				model.Read(filename);
				for (int i = 0; i < model.m_object_table.Count(); ++i) {
					const ON_Mesh *mesh = ON_Mesh::Cast(model.m_object_table[i].m_object);
					if (!mesh) continue;
					shape.Append(*mesh);

					for (int i = 0; i < shape.m_V.Count(); ++i) {
						shape.m_V[i] *= scale;
						shape.m_V[i] += ON_3fPoint(position);
					}
				}
				if (!shape.HasVertexNormals()) shape.ComputeVertexNormals();
			} else if (std::strstr(filename, ".ply") != 0) {
				ConvertFromPLY(filename, scale, position, shape);
			} else if (std::strstr(filename, ".stl") != 0){
				ON_SimpleArray<ON__UINT8> data;
				if (!ReadFile(filename, data)) continue;
				ConvertFromBinarySTL(data, scale, position, shape);
				shape.ComputeVertexNormals();
			}
		}
	}

	std::fprintf(stderr, "Constructing tree.\n");
	shapeidx2fidx.Append(0U);
	for (int k = 0; k < shapes.Count(); ++k){
		cshape.Append(shapes[k]);
		shapeidx2fidx.Append(*shapeidx2fidx.Last() + shapes[k].FaceCount());
	}

	// ގ̒`
	Materials mats((*args_p)["materials"], shapes_p, shape2matidx);
//	double ref_index = 1.1;

	// ̒`
	std::fprintf(stderr, "Defining environment.\n");
	Environment environment((*args_p)["environment"]);

	// f[^
	std::fprintf(stderr, "Defining lightsource.\n");
	LightSources light_src((*args_p)["lightsources"]);

	// J
	std::fprintf(stderr, "Definig cameras.\n");
	Cameras cameras((*args_p)["cameras"]);

	std::fprintf(stderr, "Raytracing.\n");

	// ǐ
	CommonInfo ci;
	ci.light_src = &light_src;
	ci.materials = &mats;
	ci.environment = &environment;
	ci.cnt_10 = light_src.RayCount() / 10;
	ci.ray_cursor = 0;
	ci.scene.Initialize(&cshape);
	ci.shapeidx2fidx = &shapeidx2fidx;
	ci.shape2matidx = &shape2matidx;
	ci.shape2fndm = &shape2fndm;

	{
		ON_ClassArray<ON_ClassArray<ON_3dRay> > &raies_last = ci.raies_last;
		raies_last.SetCapacity(light_src.LSCount());
		raies_last.SetCount(light_src.LSCount());
		ON_ClassArray<ON_SimpleArray<double> > &flux_last = ci.flux_last;
		flux_last.SetCapacity(light_src.LSCount());
		flux_last.SetCount(light_src.LSCount());
		for (int i = 0; i < light_src.LSCount(); ++i){
			raies_last[i].SetCapacity(light_src.RayCount(i));
			raies_last[i].SetCount(raies_last[i].Capacity());
			flux_last[i].SetCapacity(light_src.RayCount(i));
			flux_last[i].SetCount(flux_last[i].Capacity());
		}
	}

	// ̏AXbh̎擾
	size_t threads_count = mist::get_cpu_num();
	auto c1 = std::chrono::system_clock::now();

	struct Thread {
		Thread() {}
		int progress_current, progress_total;
		int thread_idx, num_threads;
		MeshRayIntersection mri;
		xorshift_rnd_32bit rnd;
		Cameras::Camera *camera;
		CommonInfo *ci;
		enum class OutputType {
			LDR, HDR
		}output_type;
		union{
			gdImagePtr ldr;
			float *exr;
		}img;
		uint64_t total_intersect_cnt, total_error_cnt;
		ON_ClassArray<ON_Polyline> pols;
		void execute() {
			int count_per_pass = camera->pass;
			total_intersect_cnt = 0;
			total_error_cnt = 0;
			progress_total = camera->pixel_height - thread_idx;
			bool error;
			int pixel_width = camera->pixel_width;
			int pixel_height = camera->pixel_height;
			for (int iy = thread_idx; iy < pixel_height; iy += num_threads) {
				progress_current = iy - thread_idx;
				for (int ix = 0; ix < pixel_width; ++ix) {
					double u = (static_cast<double>(ix) * 2.0 / static_cast<double>(pixel_width) - 1) * camera->horz_range;
					double v = (static_cast<double>(iy) * 2.0 / static_cast<double>(pixel_height) - 1) * camera->vert_range;
					ON_3dRay ray, ray_o;

					switch(camera->proj_mode){
						case Cameras::Camera::Parallel: {
							// se
							ray.m_P = camera->pln.PointAt(u, v);
							ray.m_V = camera->eye;
							break;
						}
						case Cameras::Camera::Perspective: {
							// e
							ray.m_P = camera->origin;
							ray.m_V = camera->pln.PointAt(u, v) - ray.m_P;
							ray.m_V.Unitize();
							break;
						}
					}

					double rgb[3] = { 0, 0, 0 };
					int counter_per_pass_performed = 0;
					for (int h = 0; h < count_per_pass; ++h) {
						double power[3] = { 1, 1, 1 };
						ON_Polyline pol;
						int cnt = RayTrace(ray, 1.0, mri, ci, rnd, ray_o, power, nullptr, error);
						if (error) {
							++total_error_cnt;
							continue;
						}

						total_intersect_cnt += cnt;

						auto env_rgb = (*ci->environment)(ray_o.m_V);
						rgb[0] += env_rgb.r * power[0];
						rgb[1] += env_rgb.g * power[1];
						rgb[2] += env_rgb.b * power[2];
						++counter_per_pass_performed;
						if (cnt == 0) break;
					}
					for (int h = 0; h < 3; ++h) {
						rgb[h] /= static_cast<double>(counter_per_pass_performed);
					}

					if (output_type == OutputType::LDR) {
						for (int h = 0; h < 3; ++h) {
							rgb[h] = std::pow(rgb[h], 1 / 2.2) * 255.0;
							if (rgb[h] >= 255.0) rgb[h] = 255;
						}
						img.ldr->tpixels[pixel_height - iy - 1][pixel_width - ix - 1] = gdTrueColor(static_cast<int>(rgb[0]), static_cast<int>(rgb[1]), static_cast<int>(rgb[2]));
					} else {
						float *p = &img.exr[((pixel_height-iy-1)*pixel_width + (pixel_width - ix - 1)) * 4];
						for (int h = 0; h < 3; ++h) {
							p[h] = static_cast<float>(rgb[h]);
						}
					}

				}
			}
			progress_current = progress_total;
		}
	};

	std::printf("start\n");
	xorshift_rnd_32bit rnd;
	rnd.init(444, 2531);
	ON_ClassArray<Thread> threads;
	for (int i = 0; i < threads_count; ++i) {
		Thread &th = threads.AppendNew();
		th.thread_idx = i;
		th.num_threads = threads_count;
		th.progress_total = 0;
		th.progress_current = 0;
		th.mri.Initialize(&cshape, &ci.scene.scene);
		th.rnd.init(static_cast<int>(rnd() * 10000000.0) + 1234, static_cast<int>(rnd() * 10000000.0) + 1234);
		th.ci = &ci;
		th.img.ldr = nullptr;
		th.camera = nullptr;
	}

	std::unique_ptr<thpool_, decltype(&thpool_destroy)> thpool(thpool_init(threads_count), thpool_destroy);

	for (int j = 0; j < cameras.cameras.Count(); ++j){
		auto &cmr = cameras.cameras[j];
		std::vector<float> exr;
		Thread::OutputType ot;
		gdImagePtr ldr = nullptr;
		if (cmr.output_filename.Right(4) == ".exr") {
			ot = Thread::OutputType::HDR;
			exr.resize(cmr.pixel_width * cmr.pixel_height * 4);
		} else {
			ot = Thread::OutputType::LDR;
			ldr = gdImageCreateTrueColor(cmr.pixel_width, cmr.pixel_height);
		}
		std::printf("camera # %d\n", j + 1);
		for (int i = 0; i < threads_count; ++i) {
			Thread &th = threads[i];

			th.output_type = ot;
			th.camera = &cameras.cameras[j];
			if (ot == Thread::OutputType::HDR) {
				th.img.exr = exr.data();
			} else {
				th.img.ldr = ldr;
			}
			::thpool_add_work(thpool.get(), [](void *arg) { static_cast<Thread *>(arg)->execute(); }, &th);
		}
		for (;;) {
			uint64_t progress_current = 0, progress_total = 0;
			for (int i = 0; i < threads.Count(); ++i) {
				auto &th = threads[i];
				progress_current += th.progress_current;
				progress_total += th.progress_total;
			}
			if (progress_total > 0) {
				double ratio = static_cast<double>(progress_current) * 100.0 / static_cast<double>(progress_total);
				std::printf("%lld / %lld (%5.1f %%)\n", progress_current, progress_total, ratio);
				std::fflush(stdout);
				if (progress_current == progress_total) break;
			}
			::sleep(2);
		}
		::thpool_wait(thpool.get());
		if (ldr != nullptr) {
			::gdImageFile(ldr, cmr.output_filename);
			::gdImageDestroy(ldr);
		} else if (exr.size()) {
			const char *err;
			SaveEXR(exr.data(), cmr.pixel_width, cmr.pixel_height, 4, 0, cmr.output_filename, &err);
		}
	}

	uint64_t total_intersect_cnt = 0, total_error_cnt = 0;


	auto c2 = std::chrono::system_clock::now();
	std::printf("%f msec.\n", static_cast<double>(std::chrono::duration_cast<std::chrono::microseconds>(c2 - c1).count()) / 1000.0);
	return 0;
}
